# tests to apply to all transformers
import pytest
import tubular.base as base
import tubular.capping as capping
import tubular.comparison as comparison
import tubular.dates as dates
import tubular.imputers as imputers
import tubular.mapping as mapping
import tubular.misc as misc
import tubular.nominal as nominal
import tubular.numeric as numeric
import tubular.strings as strings
import sklearn.base as b


class TestInit(object):
    """Tests for transformer.init()."""

    def ListOfTransformers():
        """List of transformers in tubular to be used in subsequent tests."""

        list_of_transformers = [
            base.BaseTransformer(columns=["a"]),
            base.DataFrameMethodTransformer(
                new_column_name="a", pd_method_name="sum", columns="b"
            ),
            capping.CappingTransformer(capping_values={"a": [0.1, 0.2]}),
            capping.OutOfRangeNullTransformer(capping_values={"a": [0.1, 0.2]}),
            comparison.EqualityChecker(columns=["a", "b"], new_col_name="c"),
            dates.DateDiffLeapYearTransformer(
                column_lower="a", column_upper="b", new_column_name="c", drop_cols=True
            ),
            dates.DateDifferenceTransformer(
                column_lower="a", column_upper="b", new_column_name="c", units="D"
            ),
            dates.ToDatetimeTransformer(column="a", new_column_name="b"),
            dates.DatetimeInfoExtractor(columns="a"),
            dates.SeriesDtMethodTransformer(
                new_column_name="a", pd_method_name="month", column="b"
            ),
            dates.BetweenDatesTransformer(
                column_lower="a",
                column_upper="b",
                column_between="c",
                new_column_name="c",
            ),
            dates.DatetimeSinusoidCalculator(
                "a",
                "sin",
                "month",
                12,
            ),
            imputers.BaseImputer(),
            imputers.ArbitraryImputer(impute_value=1, columns="a"),
            imputers.MedianImputer(columns="a"),
            imputers.MeanImputer(columns="a"),
            imputers.ModeImputer(columns="a"),
            imputers.NearestMeanResponseImputer(columns="a"),
            imputers.NullIndicator(columns="a"),
            mapping.BaseMappingTransformer(mappings={"a": {1: 2, 3: 4}}),
            mapping.BaseMappingTransformMixin(),
            mapping.MappingTransformer(mappings={"a": {1: 2, 3: 4}}),
            mapping.CrossColumnMappingTransformer(
                adjust_column="b", mappings={"a": {1: 2, 3: 4}}
            ),
            mapping.CrossColumnMultiplyTransformer(
                adjust_column="b", mappings={"a": {1: 2, 3: 4}}
            ),
            mapping.CrossColumnAddTransformer(
                adjust_column="b", mappings={"a": {1: 2, 3: 4}}
            ),
            misc.SetValueTransformer(columns="a", value=1),
            misc.SetColumnDtype(columns="a", dtype=str),
            nominal.BaseNominalTransformer(),
            nominal.NominalToIntegerTransformer(columns="a"),
            nominal.GroupRareLevelsTransformer(columns="a"),
            nominal.MeanResponseTransformer(columns="a"),
            nominal.OrdinalEncoderTransformer(columns="a"),
            nominal.OneHotEncodingTransformer(columns="a"),
            numeric.LogTransformer(columns="a"),
            numeric.CutTransformer(column="a", new_column_name="b"),
            numeric.TwoColumnOperatorTransformer(
                pd_method_name="add",
                columns=["a", "b"],
                new_column_name="c",
            ),
            numeric.ScalingTransformer(columns="a", scaler_type="standard"),
            strings.SeriesStrMethodTransformer(
                new_column_name="a",
                pd_method_name="find",
                columns="b",
                pd_method_kwargs={"sub": "a"},
            ),
            strings.StringConcatenator(columns=["a", "b"], new_column="c"),
        ]
        return list_of_transformers

    @pytest.mark.parametrize("transformer", ListOfTransformers())
    def test_print(self, transformer):
        """
        Test that transformer can be printed.
        If an error is raised in this test it will not prevent the transformer from working correctly,
        but will stop other unit tests passing.
        """

        print(transformer)

    @pytest.mark.parametrize("transformer", ListOfTransformers())
    def test_clone(self, transformer):
        """
        Test that transformer can be used in sklearn.base.clone function.
        """

        b.clone(transformer)

    @pytest.mark.parametrize("transformer", ListOfTransformers())
    def test_unexpected_kwarg(self, transformer):
        """
        Test that transformer can be used in sklearn.base.clone function.
        """

        b.clone(transformer)



from tubular._version import __version__
from tubular import base
from tubular import capping
from tubular import dates
from tubular import imputers
from tubular import mapping
from tubular import misc
from tubular import nominal
from tubular import numeric
from tubular import strings



from tubular.base import BaseTransformer
import pandas as pd


class EqualityChecker(BaseTransformer):

    """Transformer to check if two columns are equal.

    Parameters
    ----------
    columns: list
        List containing names of the two columns to check.

    new_col_name: string
        string containing the name of the new column.

    drop_original: boolean = False
        boolean representing dropping the input columns from X after checks.

    **kwargs:
        Arbitrary keyword arguments passed onto BaseTransformer.init method.

    """

    def __init__(
        self, columns: list, new_col_name: str, drop_original: bool = False, **kwargs
    ) -> None:

        super().__init__(columns=columns, **kwargs)

        if not (isinstance(columns, list)):
            raise TypeError(f"{self.classname()}: columns should be list")

        if len(columns) != 2:
            raise ValueError(
                f"{self.classname()}: This transformer works with two columns only"
            )

        if not (isinstance(new_col_name, str)):
            raise TypeError(f"{self.classname()}: new_col_name should be str")

        if not (isinstance(drop_original, bool)):
            raise TypeError(f"{self.classname()}: drop_original should be bool")

        self.new_col_name = new_col_name
        self.drop_original = drop_original

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:

        """Create a column which is populated by the boolean
        matching between two columns iterated over rows.

        Parameters
        ----------
        X : pd.DataFrame
            Data to apply mappings to.

        Returns
        -------
        X : pd.DataFrame
            Transformed input X with additional boolean column.

        """
        X = super().transform(X)

        X[self.new_col_name] = X[self.columns[0]] == X[self.columns[1]]

        if self.drop_original:

            X.drop(self.columns, axis=1, inplace=True)

        return X



"""This module contains functions that create simple datasets that are used in the tests."""

import datetime
import pandas as pd
import numpy as np


def create_series_1(n=6):
    """Create simple series of [0:n-1]."""

    df = pd.Series(np.arange(n))

    return df


def create_1_int_column_df(n=6):
    """Create single column DataFrame of [0:n-1]."""

    df = pd.DataFrame({"a": np.arange(n)})

    return df


def create_zeros_array(shape=(10, 3)):
    """Create simple 2d numpy array of zeros of given shape."""

    arr = np.zeros(shape)

    return arr


def create_numeric_df_1():
    """Example with numeric dataframe"""
    df = pd.DataFrame(
        {
            "a": [34.48, 21.71, 32.83, 1.08, 32.93, 4.74, 2.76, 75.7, 14.08, 61.31],
            "b": [12.03, 20.32, 24.12, 24.18, 68.99, 0.0, 0.0, 59.46, 11.02, 60.68],
            "c": [17.06, 12.25, 19.15, 29.73, 1.98, 8.23, 15.22, 20.59, 3.82, 39.73],
            "d": [25.94, 70.22, 72.94, 64.55, 0.41, 13.62, 30.22, 4.6, 67.13, 10.38],
            "e": [94.3, 4.18, 51.7, 16.63, 2.6, 16.57, 3.51, 30.79, 66.19, 25.44],
        }
    )
    return df


def create_df_1():
    """Create simple DataFrame with the following...

    6 rows
    2 columns;
    - a integer 1:6
    - b object a:f
    no nulls
    """

    df = pd.DataFrame({"a": [1, 2, 3, 4, 5, 6], "b": ["a", "b", "c", "d", "e", "f"]})

    return df


def create_df_2():
    """Create simple DataFrame to use in other tests"""

    df = pd.DataFrame(
        {
            "a": [1, 2, 3, 4, 5, 6, np.NaN],
            "b": ["a", "b", "c", "d", "e", "f", np.NaN],
            "c": ["a", "b", "c", "d", "e", "f", np.NaN],
        }
    )

    df["c"] = df["c"].astype("category")

    return df


def create_df_4():
    """Create simple DataFrame to use in other tests"""

    df = pd.DataFrame(
        {
            "a": [1, 2, 3, 4, 5, 6, 7, np.NaN],
            "b": ["a", "b", "c", "d", "e", "f", "g", np.NaN],
            "c": ["a", "b", "c", "d", "e", "f", "g", np.NaN],
        }
    )

    df["c"] = df["c"].astype("category")

    return df


def create_df_3():
    """Create simple DataFrame to use in other tests"""

    df = pd.DataFrame(
        {
            "a": [1, 2, 3, 4, 5, 6, np.NaN],
            "b": [1, 2, 3, np.NaN, 7, 8, 9],
            "c": [np.NaN, 1, 2, 3, -4, -5, -6],
        }
    )

    return df


def create_df_5():
    """Create simple DataFrame to use in other tests"""

    df = pd.DataFrame(
        {
            "a": [1, 2, 3, 4, 5, 6, 7, 8, 9, np.NaN],
            "b": ["a", "a", "a", "d", "e", "f", "g", np.NaN, np.NaN, np.NaN],
            "c": ["a", "a", "c", "c", "e", "e", "f", "g", "h", np.NaN],
        }
    )

    df["c"] = df["c"].astype("category")

    return df


def create_df_6():
    """Nulls in different positions to check summing weights by col with nulls"""

    df = pd.DataFrame(
        {
            "a": [2, 2, 2, 2, np.NaN, 2, 2, 2, 3, 3],
            "b": ["a", "a", "a", "d", "e", "f", "g", np.NaN, np.NaN, np.NaN],
            "c": ["a", "b", "c", "d", "f", "f", "f", "g", "g", np.NaN],
        }
    )

    df["c"] = df["c"].astype("category")

    return df


def create_df_7():
    """Create simple DataFrame to use in other tests"""

    df = pd.DataFrame(
        {
            "a": [4, 2, 2, 1, 3],
            "b": ["x", "z", "y", "x", "x"],
            "c": ["c", "a", "a", "c", "b"],
        }
    )

    df["c"] = df["c"].astype("category")

    return df


def create_df_8():
    """Create simple DataFrame to use in other tests"""

    df = pd.DataFrame(
        {
            "a": [1, 5, 2, 3, 3],
            "b": ["w", "w", "z", "y", "x"],
            "c": ["a", "a", "c", "b", "a"],
        },
        index=[10, 15, 200, 251, 59],
    )

    df["c"] = df["c"].astype("category")

    return df


def create_df_9():
    """Create simple DataFrame to use in other tests"""

    df = pd.DataFrame(
        {
            "a": [1, 2, np.nan, 4, np.nan, 6],
            "b": [np.nan, 5, 4, 3, 2, 1],
            "c": [3, 2, 1, 4, 5, 6],
        }
    )

    return df


def create_df_10():
    """Create simple DataFrame to use in other tests"""

    df = pd.DataFrame(
        {
            "a": [1, 2, 3, 4, 5, 6, 7, 8, 9, np.NaN],
            "b": ["a", "a", "a", "d", "e", "f", "g", np.NaN, np.NaN, np.NaN],
            "c": [1, 1, 3, 4, 5, 6, 5, 8, 9, 50],
        }
    )

    return df


def create_df_11():
    """Create simple DataFrame to use in other tests"""

    df = pd.DataFrame(
        {
            "a": [1, 2, 3],
            "b": [4, 5, 6],
        }
    )

    return df


def create_large_null_df(n_col=1000):
    """Create large single row df with all null values.

    Parameters
    ----------
    n_col : int
        Number of columns to add to output dataframe. Columns are named 'col_0', 'col_1', etc..

    Returns
    -------
    data_df : pd.DataFrame
        Single row dataframe with n_col columns; entirely populated with null values.

    """

    data_dict = {}

    for i in range(n_col):

        data_dict["col_" + str(i)] = np.NaN

    data_df = pd.DataFrame(data_dict, index=[0])

    return data_df


def create_large_half_null_df(n_col=1000):
    """Create large 2 row df with all null values in 1 row and all 1s in the other.

    Parameters
    ----------
    n_col : int
        Number of columns to add to output dataframe. Columns are named 'col_0', 'col_1', etc..

    Returns
    -------
    data_df : pd.DataFrame
        Two row dataframe with n_col columns; the first row taking entirely values of 1 and the
        second taking entirely null values.

    """

    data_dict = {}

    col_values = [1.0, np.NaN]

    for i in range(n_col):

        data_dict["col_" + str(i)] = col_values

    data_df = pd.DataFrame(data_dict, index=[0, 1])

    return data_df


def create_weighted_imputers_test_df():
    """Create DataFrame to use imputer tests that correct values are imputed for weighted dataframes

    weight column contains the weights between 0 and 1
    """

    df = pd.DataFrame(
        {
            "a": [1.0, 1.0, 1.0, 3.0, 5.0, 5.0],
            "b": ["a", "a", "a", "d", "e", "e"],
            "c": ["a", "a", np.nan, np.nan, np.nan, "f"],
            "d": [1.0, 5.0, 3.0, np.nan, np.nan, 1.0],
            "response": [0, 1, 0, 1, 1, 1],
            "weight": [0.1, 0.1, 0.8, 0.5, 0.9, 0.8],
        }
    )

    return df


def create_MeanResponseTransformer_test_df():
    """Create DataFrame to use MeanResponseTransformer tests that correct values are

    DataFrame column a is the response, the other columns are categorical columns
    of types; object, category, int, float, bool.

    """

    df = pd.DataFrame(
        {
            "a": [1.0, 2.0, 3.0, 4.0, 5.0, 6.0],
            "b": ["a", "b", "c", "d", "e", "f"],
            "c": ["a", "b", "c", "d", "e", "f"],
            "d": [1, 2, 3, 4, 5, 6],
            "e": [1, 2, 3, 4, 5, 6.0],
            "f": [False, False, False, True, True, True],
            "multi_level_response": [
                "blue",
                "blue",
                "yellow",
                "yellow",
                "green",
                "green",
            ],
        }
    )

    df["c"] = df["c"].astype("category")

    return df


def create_MeanResponseTransformer_test_df_unseen_levels():
    """Create DataFrame to use in MeanResponseTransformer tests that check correct values are
    generated when using transform method on data with unseen levels.
    DataFrame column a is the response, the other columns are categorical columns
    of types; object, category, int, float, bool.

    """
    df = pd.DataFrame(
        {
            "a": [1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 3.0],
            "b": ["a", "b", "c", "d", "e", "f", "g", "h"],
            "c": ["a", "b", "c", "d", "e", "f", "g", "h"],
            "d": [1, 2, 3, 4, 5, 6, 7, 8],
            "e": [1, 2, 3, 4, 5, 6.0, 7, 8],
            "f": [False, False, False, True, True, True, True, False],
            "multi_level_response": [
                "blue",
                "blue",
                "yellow",
                "yellow",
                "green",
                "green",
                "yellow",
                "blue",
            ],
        }
    )

    df["c"] = df["c"].astype("category")

    return df


def create_OrdinalEncoderTransformer_test_df():
    """Create DataFrame to use OrdinalEncoderTransformer tests that correct values are

    DataFrame column a is the response, the other columns are categorical columns
    of types; object, category, int, float, bool.

    """

    df = pd.DataFrame(
        {
            "a": [1, 2, 3, 4, 5, 6],
            "b": ["a", "b", "c", "d", "e", "f"],
            "c": ["a", "b", "c", "d", "e", "f"],
            "d": [1, 2, 3, 4, 5, 6],
            "e": [3, 4, 5, 6, 7, 8.0],
            "f": [False, False, False, True, True, True],
        }
    )

    df["c"] = df["c"].astype("category")

    return df


def create_NearestMeanResponseImputer_test_df():
    """Create DataFrame to use in NearestMeanResponseImputer tests.

    DataFrame column c is the response, the other columns are numerical columns containing null entries.

    """

    df = pd.DataFrame(
        {
            "a": [1, 1, 2, 3, 3, np.nan],
            "b": [np.nan, np.nan, 1, 3, 3, 4],
            "c": [2, 3, 2, 1, 4, 1],
        }
    )

    return df


def create_values_map(df):

    value_map = {}

    for i in df.columns:

        value_map[i] = {}

        for j in df[i].unique():

            value_map[i][j] = j

    return value_map


def create_date_test_df():
    """Create DataFrame for DateDiffLeapYearTransformer tests."""

    df = pd.DataFrame(
        {
            "a": [
                datetime.date(1993, 9, 27),
                datetime.date(2000, 3, 19),
                datetime.date(2018, 11, 10),
                datetime.date(2018, 10, 10),
                datetime.date(2018, 10, 10),
                datetime.date(2018, 10, 10),
                datetime.date(2018, 12, 10),
                datetime.date(1985, 7, 23),
            ],
            "b": [
                datetime.date(2020, 5, 1),
                datetime.date(2019, 12, 25),
                datetime.date(2018, 11, 10),
                datetime.date(2018, 11, 10),
                datetime.date(2018, 9, 10),
                datetime.date(2015, 11, 10),
                datetime.date(2015, 11, 10),
                datetime.date(2015, 7, 23),
            ],
        }
    )

    return df


def create_date_test_nulls_df():
    """Create DataFrame with nulls only for DateDiffLeapYearTransformer, DateDifferenceTransformer tests."""

    df = pd.DataFrame(
        {
            "a": [
                np.NaN,
            ],
            "b": [
                np.NaN,
            ],
        },
        index=[0],
    )

    return df


def create_datediff_test_df():
    """Create DataFrame for DateDifferenceTransformer tests."""

    df = pd.DataFrame(
        {
            "a": [
                datetime.datetime(1993, 9, 27, 11, 58, 58),
                datetime.datetime(2000, 3, 19, 12, 59, 59),
                datetime.datetime(2018, 11, 10, 11, 59, 59),
                datetime.datetime(2018, 10, 10, 11, 59, 59),
                datetime.datetime(2018, 10, 10, 11, 59, 59),
                datetime.datetime(2018, 10, 10, 10, 59, 59),
                datetime.datetime(2018, 12, 10, 11, 59, 59),
                datetime.datetime(1985, 7, 23, 11, 59, 59),
            ],
            "b": [
                datetime.datetime(2020, 5, 1, 12, 59, 59),
                datetime.datetime(2019, 12, 25, 11, 58, 58),
                datetime.datetime(2018, 11, 10, 11, 59, 59),
                datetime.datetime(2018, 11, 10, 11, 59, 59),
                datetime.datetime(2018, 9, 10, 9, 59, 59),
                datetime.datetime(2015, 11, 10, 11, 59, 59),
                datetime.datetime(2015, 11, 10, 12, 59, 59),
                datetime.datetime(2015, 7, 23, 11, 59, 59),
            ],
        }
    )

    return df


def create_datediff_test_nulls_df():
    """Create DataFrame with nulls only for DateDifferenceTransformer tests."""

    df = pd.DataFrame(
        {
            "a": [
                datetime.datetime(1993, 9, 27, 11, 58, 58),
                np.NaN,
            ],
            "b": [
                np.NaN,
                datetime.datetime(2019, 12, 25, 11, 58, 58),
            ],
        },
        index=[0, 1],
    )

    return df


def create_to_datetime_test_df():
    """Create DataFrame to be used in the ToDatetimeTransformer tests."""

    df = pd.DataFrame(
        {"a": [1950, 1960, 2000, 2001, np.NaN, 2010], "b": [1, 2, 3, 4, 5, np.NaN]}
    )

    return df


def create_is_between_dates_df_1():
    """Create df to use in IsBetweenDates tests. Contains 3 columns of 2 datatime values."""

    df = pd.DataFrame(
        {
            "a": pd.date_range(start="1/1/2016", end="27/02/2017", periods=2),
            "b": pd.date_range(start="1/2/2016", end="27/09/2017", periods=2),
            "c": pd.date_range(start="1/3/2016", end="27/04/2017", periods=2),
        }
    )

    return df


def create_is_between_dates_df_2():
    """Create df to use in IsBetweenDates tests. Contains 3 columns of 5 datatime values, covers edge cases."""

    df = pd.DataFrame(
        {
            "a": [
                datetime.datetime(1990, 2, 1),
                datetime.datetime(1990, 2, 1),
                datetime.datetime(1990, 2, 1),
                datetime.datetime(1990, 2, 1),
                datetime.datetime(1990, 2, 1),
                datetime.datetime(1990, 2, 1),
            ],
            "b": [
                datetime.datetime(1990, 1, 20),
                datetime.datetime(1990, 2, 1),
                datetime.datetime(1990, 2, 2),
                datetime.datetime(1990, 2, 6),
                datetime.datetime(1990, 3, 1),
                datetime.datetime(1990, 3, 2),
            ],
            "c": [
                datetime.datetime(1990, 3, 1),
                datetime.datetime(1990, 3, 1),
                datetime.datetime(1990, 3, 1),
                datetime.datetime(1990, 3, 1),
                datetime.datetime(1990, 3, 1),
                datetime.datetime(1990, 3, 1),
            ],
        }
    )

    return df


# Example DataFrame for downcasting dtypes tests
def create_downcast_df():
    """Create a dataframe with mixed dtypes to use in downcasting tests."""
    df = pd.DataFrame(
        {
            "a": [1, 2, 3, 4, 5],
            "b": [1.0, 2.0, 3.0, 4.0, 5.0],
        }
    )

    return df



"""
This module contains a transformer that applies capping to numeric columns.
"""

import pandas as pd
import numpy as np
import warnings
import copy

from tubular.base import BaseTransformer


class CappingTransformer(BaseTransformer):
    """Transformer to cap numeric values at both or either minimum and maximum values.

    For max capping any values above the cap value will be set to the cap. Similarly for min capping
    any values below the cap will be set to the cap. Only works for numeric columns.

    Parameters
    ----------
    capping_values : dict or None, default = None
        Dictionary of capping values to apply to each column. The keys in the dict should be the
        column names and each item in the dict should be a list of length 2. Items in the lists
        should be ints or floats or None. The first item in the list is the minimum capping value
        and the second item in the list is the maximum capping value. If None is supplied for
        either value then that capping will not take place for that particular column. Both items
        in the lists cannot be None. Either one of capping_values or quantiles must be supplied.

    quantiles : dict or None, default = None
        Dictionary of quantiles in the range [0, 1] to set capping values at for each column.
        The keys in the dict should be the column names and each item in the dict should be a
        list of length 2. Items in the lists should be ints or floats or None. The first item in the
        list is the lower quantile and the second item is the upper quantile to set the capping
        value from. The fit method calculates the values quantile from the input data X. If None is
        supplied for either value then that capping will not take place for that particular column.
        Both items in the lists cannot be None. Either one of capping_values or quantiles must be
        supplied.

    weights_column : str or None, default = None
        Optional weights column argument that can be used in combination with quantiles. Not used
        if capping_values is supplied. Allows weighted quantiles to be calculated.

    **kwargs
        Arbitrary keyword arguments passed onto BaseTransformer.init method.

    Attributes
    ----------
    capping_values : dict or None
        Capping values to apply to each column, capping_values argument.

    quantiles : dict or None
        Quantiles to set capping values at from input data. Will be empty after init, values
        populated when fit is run.

    weights_column : str or None
        weights_column argument.

    _replacement_values : dict
        Replacement values when capping is applied. Will be a copy of capping_values.

    """

    def __init__(
        self, capping_values=None, quantiles=None, weights_column=None, **kwargs
    ):

        if capping_values is None and quantiles is None:

            raise ValueError(
                f"{self.classname()}: both capping_values and quantiles are None, either supply capping values in the "
                "capping_values argument or supply quantiles that can be learnt in the fit method"
            )

        if capping_values is not None and quantiles is not None:

            raise ValueError(
                f"{self.classname()}: both capping_values and quantiles are not None, supply one or the other"
            )

        if capping_values is not None:

            self.check_capping_values_dict(capping_values, "capping_values")

            self.capping_values = capping_values

            super().__init__(columns=list(capping_values.keys()), **kwargs)

        if quantiles is not None:

            self.check_capping_values_dict(quantiles, "quantiles")

            for k, quantile_values in quantiles.items():

                for quantile_value in quantile_values:

                    if quantile_value is not None:

                        if quantile_value < 0 or quantile_value > 1:

                            raise ValueError(
                                f"{self.classname()}: quantile values must be in the range [0, 1] but got {quantile_value} for key {k}"
                            )

            self.capping_values = {}

            super().__init__(columns=list(quantiles.keys()), **kwargs)

        self.quantiles = quantiles
        self.weights_column = weights_column
        self._replacement_values = copy.deepcopy(self.capping_values)

    def check_capping_values_dict(self, capping_values_dict, dict_name):
        """Performs checks on a dictionary passed to ."""

        if type(capping_values_dict) is not dict:

            raise TypeError(
                f"{self.classname()}: {dict_name} should be dict of columns and capping values"
            )

        for k, cap_values in capping_values_dict.items():

            if type(k) is not str:

                raise TypeError(
                    f"{self.classname()}: all keys in {dict_name} should be str, but got {type(k)}"
                )

            if type(cap_values) is not list:

                raise TypeError(
                    f"{self.classname()}: each item in {dict_name} should be a list, but got {type(cap_values)} for key {k}"
                )

            if len(cap_values) != 2:

                raise ValueError(
                    f"{self.classname()}: each item in {dict_name} should be length 2, but got {len(cap_values)} for key {k}"
                )

            for cap_value in cap_values:

                if cap_value is not None:

                    if type(cap_value) not in [int, float]:

                        raise TypeError(
                            f"{self.classname()}: each item in {dict_name} lists must contain numeric values or None, got {type(cap_value)} for key {k}"
                        )

                    if np.isnan(cap_value) or np.isinf(cap_value):

                        raise ValueError(
                            f"{self.classname()}: item in {dict_name} lists contains numpy NaN or Inf values"
                        )

            if all([cap_value is not None for cap_value in cap_values]):

                if cap_values[0] >= cap_values[1]:

                    raise ValueError(
                        f"{self.classname()}: lower value is greater than or equal to upper value for key {k}"
                    )

            if all([cap_value is None for cap_value in cap_values]):

                raise ValueError(
                    f"{self.classname()}: both values are None for key {k}"
                )

    def fit(self, X, y=None):
        """Learn capping values from input data X.

        Calculates the quantiles to cap at given the quantiles dictionary supplied
        when initialising the transformer. Saves learnt values in the capping_values
        attribute.

        Parameters
        ----------
        X : pd.DataFrame
            A dataframe with required columns to be capped.

        y : None
            Required for pipeline.

        """

        super().fit(X, y)

        if self.quantiles is not None:

            for col in self.columns:

                if self.weights_column is None:

                    cap_values = self.prepare_quantiles(
                        X[col], self.quantiles[col], self.weights_column
                    )

                else:

                    cap_values = self.prepare_quantiles(
                        X[col], self.quantiles[col], X[self.weights_column]
                    )

                self.capping_values[col] = cap_values

        else:

            warnings.warn(
                f"{self.classname()}: quantiles not set so no fitting done in CappingTransformer"
            )

        self._replacement_values = copy.deepcopy(self.capping_values)

        return self

    def prepare_quantiles(self, values, quantiles, sample_weight=None):
        """Method to call the weighted_quantile method and prepare the outputs.

        If there are no None values in the supplied quantiles then the outputs from weighted_quantile
        are returned as is. If there are then prepare_quantiles removes the None values before
        calling weighted_quantile and adds them back into the output, in the same position, after
        calling.

        Parameters
        ----------
        values : pd.Series or np.array
            A dataframe column with values to calculate quantiles from.

        quantiles : None
            Weighted quantiles to calculate. Must all be between 0 and 1.

        sample_weight : pd.Series or np.array or None, default = None
            Sample weights for each item in values, must be the same lenght as values. If
            not supplied then unit weights will be used.

        Returns
        -------
        interp_quantiles : list
            List containing computed quantiles.

        """

        if quantiles[0] is None:

            quantiles = np.array([quantiles[1]])

            results_no_none = self.weighted_quantile(values, quantiles, sample_weight)

            results = [None] + results_no_none

        elif quantiles[1] is None:

            quantiles = np.array([quantiles[0]])

            results_no_none = self.weighted_quantile(values, quantiles, sample_weight)

            results = results_no_none + [None]

        else:

            results = self.weighted_quantile(values, quantiles, sample_weight)

        return results

    def weighted_quantile(self, values, quantiles, sample_weight=None):
        """Method to calculate weighted quantiles.

        This method is adapted from the "Completely vectorized numpy solution" answer from user
        Alleo (https://stackoverflow.com/users/498892/alleo) to the following stackoverflow question;
        https://stackoverflow.com/questions/21844024/weighted-percentile-using-numpy. This
        method is also licenced under the CC-BY-SA terms, as the original code sample posted
        to stackoverflow (pre February 1, 2016) was.

        Method is similar to numpy.percentile, but supports weights. Supplied quantiles should be
        in the range [0, 1]. Method calculates cumulative % of weight for each observation,
        then interpolates between these observations to calculate the desired quantiles. Null values
        in the observations (values) and 0 weight observations are filtered out before
        calculating.

        Parameters
        ----------
        values : pd.Series or np.array
            A dataframe column with values to calculate quantiles from.

        quantiles : None
            Weighted quantiles to calculate. Must all be between 0 and 1.

        sample_weight : pd.Series or np.array or None, default = None
            Sample weights for each item in values, must be the same lenght as values. If
            not supplied then unit weights will be used.

        Returns
        -------
        interp_quantiles : list
            List containing computed quantiles.

        Examples
        --------
        >>> x = CappingTransformer(capping_values={"a": [2, 10]})
        >>> quantiles_to_compute = [0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0]
        >>> computed_quantiles = x.weighted_quantile(values = [1, 2, 3], sample_weight = [1, 1, 1], quantiles = quantiles_to_compute)
        >>> [round(q, 1) for q in computed_quantiles]
        [1.0, 1.0, 1.0, 1.0, 1.2, 1.5, 1.8, 2.1, 2.4, 2.7, 3.0]
        >>>
        >>> computed_quantiles = x.weighted_quantile(values = [1, 2, 3], sample_weight = [0, 1, 0], quantiles = quantiles_to_compute)
        >>> [round(q, 1) for q in computed_quantiles]
        [2.0, 2.0, 2.0, 2.0, 2.0, 2.0, 2.0, 2.0, 2.0, 2.0, 2.0]
        >>>
        >>> computed_quantiles = x.weighted_quantile(values = [1, 2, 3], sample_weight = [1, 1, 0], quantiles = quantiles_to_compute)
        >>> [round(q, 1) for q in computed_quantiles]
        [1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.2, 1.4, 1.6, 1.8, 2.0]
        >>>
        >>> computed_quantiles = x.weighted_quantile(values = [1, 2, 3, 4, 5], sample_weight = [1, 1, 1, 1, 1], quantiles = quantiles_to_compute)
        >>> [round(q, 1) for q in computed_quantiles]
        [1.0, 1.0, 1.0, 1.5, 2.0, 2.5, 3.0, 3.5, 4.0, 4.5, 5.0]
        >>>
        >>> computed_quantiles = x.weighted_quantile(values = [1, 2, 3, 4, 5], sample_weight = [1, 0, 1, 0, 1], quantiles = [0, 0.5, 1.0])
        >>> [round(q, 1) for q in computed_quantiles]
        [1.0, 2.0, 5.0]

        """

        if sample_weight is None:
            sample_weight = np.ones(len(values))
        else:
            sample_weight = np.array(sample_weight)

        if np.isnan(sample_weight).sum() > 0:
            raise ValueError(f"{self.classname()}: null values in sample weights")

        if np.isinf(sample_weight).sum() > 0:
            raise ValueError(f"{self.classname()}: infinite values in sample weights")

        if (sample_weight < 0).sum() > 0:
            raise ValueError(f"{self.classname()}: negative weights in sample weights")

        if sample_weight.sum() <= 0:
            raise ValueError(
                f"{self.classname()}: total sample weights are not greater than 0"
            )

        values = np.array(values)
        quantiles = np.array(quantiles)

        nan_filter = ~np.isnan(values)
        values = values[nan_filter]
        sample_weight = sample_weight[nan_filter]

        zero_weight_filter = ~(sample_weight == 0)
        values = values[zero_weight_filter]
        sample_weight = sample_weight[zero_weight_filter]

        sorter = np.argsort(values, kind="stable")
        values = values[sorter]
        sample_weight = sample_weight[sorter]

        weighted_quantiles = np.cumsum(sample_weight)
        weighted_quantiles = weighted_quantiles / np.sum(sample_weight)

        interp_quantiles = list(np.interp(quantiles, weighted_quantiles, values))

        return interp_quantiles

    def transform(self, X):
        """Apply capping to columns in X.

        If cap_value_max is set, any values above cap_value_max will be set to cap_value_max. If cap_value_min
        is set any values below cap_value_min will be set to cap_value_min. Only works or numeric columns.

        Parameters
        ----------
        X : pd.DataFrame
            Data to apply capping to.

        Returns
        -------
        X : pd.DataFrame
            Transformed input X with min and max capping applied to the specified columns.

        """

        self.check_is_fitted(["capping_values"])
        self.check_is_fitted(["_replacement_values"])

        if self.capping_values == {}:

            raise ValueError(
                f"{self.classname()}: capping_values attribute is an empty dict - perhaps the fit method has not been run yet"
            )

        if self._replacement_values == {}:

            raise ValueError(
                f"{self.classname()}: _replacement_values attribute is an empty dict - perhaps the fit method has not been run yet"
            )

        X = super().transform(X)

        numeric_column_types = X[self.columns].apply(
            pd.api.types.is_numeric_dtype, axis=0
        )

        if not numeric_column_types.all():

            non_numeric_columns = list(
                numeric_column_types.loc[~numeric_column_types].index
            )

            raise TypeError(
                f"{self.classname()}: The following columns are not numeric in X; {non_numeric_columns}"
            )

        for col in self.columns:

            cap_value_min = self.capping_values[col][0]
            cap_value_max = self.capping_values[col][1]

            replacement_min = self._replacement_values[col][0]
            replacement_max = self._replacement_values[col][1]

            if cap_value_min is not None:

                X.loc[X[col] < cap_value_min, col] = replacement_min

            if cap_value_max is not None:

                X.loc[X[col] > cap_value_max, col] = replacement_max

        return X


class OutOfRangeNullTransformer(CappingTransformer):
    """Transformer to set values outside of a range to null.

    This transformer sets the cut off values in the same way as
    the CappingTransformer. So either the user can specify them
    directly in the capping_values argument or they can be calculated
    in the fit method, if the user supplies the quantiles argument.

    Parameters
    ----------
    capping_values : dict or None, default = None
        Dictionary of capping values to apply to each column. The keys in the dict should be the
        column names and each item in the dict should be a list of length 2. Items in the lists
        should be ints or floats or None. The first item in the list is the minimum capping value
        and the second item in the list is the maximum capping value. If None is supplied for
        either value then that capping will not take place for that particular column. Both items
        in the lists cannot be None. Either one of capping_values or quantiles must be supplied.

    quantiles : dict or None, default = None
        Dictionary of quantiles to set capping values at for each column. The keys in the dict
        should be the column names and each item in the dict should be a list of length 2. Items
        in the lists should be ints or floats or None. The first item in the list is the lower
        quantile and the second item is the upper quantile to set the capping value from. The fit
        method calculates the values quantile from the input data X. If None is supplied for
        either value then that capping will not take place for that particular column. Both items
        in the lists cannot be None. Either one of capping_values or quantiles must be supplied.

    weights_column : str or None, default = None
        Optional weights column argument that can be used in combination with quantiles. Not used
        if capping_values is supplied. Allows weighted quantiles to be calculated.

    **kwargs
        Arbitrary keyword arguments passed onto BaseTransformer.init method.

    Attributes
    ----------
    capping_values : dict or None
        Capping values to apply to each column, capping_values argument.

    quantiles : dict or None
        Quantiles to set capping values at from input data. Will be empty after init, values
        populated when fit is run.

    weights_column : str or None
        weights_column argument.

    _replacement_values : dict
        Replacement values when capping is applied. This will contain nulls for each column.

    """

    def __init__(
        self, capping_values=None, quantiles=None, weights_column=None, **kwargs
    ):

        super().__init__(
            capping_values=capping_values,
            quantiles=quantiles,
            weights_column=weights_column,
            **kwargs,
        )

        self.set_replacement_values()

    def set_replacement_values(self):
        """Method to set the _replacement_values to have all null values.

        Keeps the existing keys in the _replacement_values dict and sets all values (except None) in the lists to np.NaN. Any None
        values remain in place.
        """

        for k, replacements_list in self._replacement_values.items():

            null_replacements_list = [
                np.NaN if replace_value is not None else None
                for replace_value in replacements_list
            ]

            self._replacement_values[k] = null_replacements_list

    def fit(self, X, y=None):
        """Learn capping values from input data X.

        Calculates the quantiles to cap at given the quantiles dictionary supplied
        when initialising the transformer. Saves learnt values in the capping_values
        attribute.

        Parameters
        ----------
        X : pd.DataFrame
            A dataframe with required columns to be capped.

        y : None
            Required for pipeline.

        """

        super().fit(X=X, y=y)

        self.set_replacement_values()

        return self



"""
This module contains transformers that apply different types of mappings to columns.
"""

import pandas as pd
from pandas.api.types import is_categorical_dtype
import numpy as np
from collections import OrderedDict
import warnings


from tubular.base import BaseTransformer, ReturnKeyDict


class BaseMappingTransformer(BaseTransformer):
    """Base Transformer Extension for mapping transformers.

    Parameters
    ----------
    mappings : dict
        Dictionary containing column mappings. Each value in mappings should be a dictionary
        of key (column to apply mapping to) value (mapping dict for given columns) pairs. For
        example the following dict {'a': {1: 2, 3: 4}, 'b': {'a': 1, 'b': 2}} would specify
        a mapping for column a of 1->2, 3->4 and a mapping for column b of 'a'->1, b->2.

    **kwargs
        Arbitrary keyword arguments passed onto BaseTransformer.init method.

    Attributes
    ----------
    mappings : dict
        Dictionary of mappings for each column individually. The dict passed to mappings in
        init is set to the mappings attribute.

    """

    def __init__(self, mappings, **kwargs):

        if isinstance(mappings, dict):

            if not len(mappings) > 0:

                raise ValueError(f"{self.classname()}: mappings has no values")

            for j in mappings.values():

                if not isinstance(j, dict):

                    raise ValueError(
                        f"{self.classname()}: values in mappings dictionary should be dictionaries"
                    )

            self.mappings = mappings

        else:

            raise ValueError(f"{self.classname()}: mappings must be a dictionary")

        columns = list(mappings.keys())

        super().__init__(columns=columns, **kwargs)

    def transform(self, X):
        """Base mapping transformer transform method.  Checks that the mappings
        dict has been fitted and calls the BaseTransformer transform method.

        Parameters
        ----------
        X : pd.DataFrame
            Data to apply mappings to.

        Returns
        -------
        X : pd.DataFrame
            Input X, copied if specified by user.

        """

        self.check_is_fitted(["mappings"])

        X = super().transform(X)

        return X


class BaseMappingTransformMixin(BaseTransformer):
    """Mixin class to apply standard pd.Series.map transform method.

    Transformer uses the mappings attribute which should be a dict of dicts/mappings
    for each required column.

    """

    def transform(self, X):
        """Applies the mapping defined in the mappings dict to each column in the columns
        attribute.

        Parameters
        ----------
        X : pd.DataFrame
            Data with nominal columns to transform.

        Returns
        -------
        X : pd.DataFrame
            Transformed input X with levels mapped accoriding to mappings dict.

        """

        self.check_is_fitted(["mappings"])

        X = super().transform(X)

        for c in self.columns:

            X[c] = X[c].map(self.mappings[c])

        return X


class MappingTransformer(BaseMappingTransformer, BaseMappingTransformMixin):
    """Transformer to map values in columns to other values e.g. to merge two levels into one.

    Note, the MappingTransformer does not require 'self-mappings' to be defined i.e. if you want
    to map a value to itself, you can omit this value from the mappings rather than having to
    map it to itself. This is because it uses the ReturnKeyDict  type to store the mappings
    for each columns, this dict will return the key i.e. the original value in that row if it
    is not available in the mapping dict.

    This transformer inherits from BaseMappingTransformMixin as well as the BaseMappingTransformer
    in order to access the standard pd.Series.map transform function.

    Parameters
    ----------
    mappings : dict
        Dictionary containing column mappings. Each value in mappings should be a dictionary
        of key (column to apply mapping to) value (mapping dict for given columns) pairs. For
        example the following dict {'a': {1: 2, 3: 4}, 'b': {'a': 1, 'b': 2}} would specify
        a mapping for column a of 1->2, 3->4 and a mapping for column b of 'a'->1, b->2.

    **kwargs
        Arbitrary keyword arguments passed onto BaseMappingTransformer.init method.

    Attributes
    ----------
    mappings : dict
        Dictionary of mappings for each column individually. The dict passed to mappings in
        init is set to the mappings attribute.

    """

    def __init__(self, mappings, **kwargs):

        for k, v in mappings.items():

            if isinstance(v, dict):

                mappings[k] = ReturnKeyDict(v)

            else:

                raise TypeError(
                    f"{self.classname()}: each item in mappings should be a dict but got type {type(v)} for key {k}"
                )

        BaseMappingTransformer.__init__(self, mappings=mappings, **kwargs)

    def transform(self, X, suppress_dtype_warning=False):
        """Transform the input data X according to the mappings in the mappings attribute dict.

        This method calls the BaseMappingTransformMixin.transform. Note, this transform method is
        different to some of the transform methods in the nominal module, even though they also
        use the BaseMappingTransformMixin.transform method. Here, if a value does not exist in
        the mapping it is unchanged.

        Due to the way pd.Series.map works, mappings can result in column dtypes changing,
        sometimes unexpectedly. If the result of the mappings is a dtype that doesn't match
        the original dtype, or the dtype of the values provided in the mapping a warning
        will be raised. This normally results from an incomplete mapping being provided,
        or a mix of dtypes causing pandas to default to the object dtype.

        For columns with a 'category' dtype the warning will not be raised.

        Parameters
        ----------
        X : pd.DataFrame
            Data with nominal columns to transform.

        suppress_dtype_warning: Bool, default = False
            Whether to suppress warnings about dtype changes

        Returns
        -------
        X : pd.DataFrame
            Transformed input X with levels mapped accoriding to mappings dict.

        """
        mapped_columns = self.mappings.keys()
        original_dtypes = X[mapped_columns].dtypes

        for col in mapped_columns:

            values_to_be_mapped = set(self.mappings[col].keys())
            values_in_df = set(X[col].unique())

            if len(values_to_be_mapped.intersection(values_in_df)) == 0:

                warnings.warn(
                    f"{self.classname()}: No values from mapping for {col} exist in dataframe."
                )

            if len(values_to_be_mapped.difference(values_in_df)) > 0:

                warnings.warn(
                    f"{self.classname()}: There are values in the mapping for {col} that are not present in the dataframe"
                )

        X = BaseMappingTransformMixin.transform(self, X)

        mapped_dtypes = X[mapped_columns].dtypes

        if not suppress_dtype_warning:

            for col in mapped_columns:

                col_mappings = pd.Series(self.mappings[col])
                mapping_dtype = col_mappings.dtype

                if (mapped_dtypes[col] != mapping_dtype) and (
                    mapped_dtypes[col] != original_dtypes[col]
                ):

                    # Confirm the initial and end dtypes are not categories
                    if not (
                        is_categorical_dtype(original_dtypes[col])
                        and is_categorical_dtype(mapped_dtypes[col])
                    ):

                        warnings.warn(
                            f"{self.classname()}: This mapping changes {col} dtype from {original_dtypes[col]} to {mapped_dtypes[col]}. This is often caused by having multiple dtypes in one column, or by not mapping all values."
                        )

        return X


class CrossColumnMappingTransformer(BaseMappingTransformer):
    """Transformer to adjust values in one column based on the values of another column.

    Parameters
    ----------
    adjust_column : str
        The column to be adjusted.

    mappings : dict or OrderedDict
        Dictionary containing adjustments. Each value in adjustments should be a dictionary
        of key (column to apply adjustment based on) value (adjustment dict for given columns) pairs. For
        example the following dict {'a': {1: 'a', 3: 'b'}, 'b': {'a': 1, 'b': 2}}
        would replace the values in the adjustment column based off the values in column a using the mapping
        1->'a', 3->'b' and also replace based off the values in column b using a mapping 'a'->1, 'b'->2.
        If more than one column is defined for this mapping, then this object must be an OrderedDict
        to ensure reproducibility.

    **kwargs
        Arbitrary keyword arguments passed onto BaseTransformer.init method.

    Attributes
    ----------

    adjust_column : str
        Column containing the values to be adjusted.

    mappings : dict
        Dictionary of mappings for each column individually to be applied to the adjust_column.
        The dict passed to mappings in init is set to the mappings attribute.


    """

    def __init__(self, adjust_column, mappings, **kwargs):

        super().__init__(mappings=mappings, **kwargs)

        if not isinstance(adjust_column, str):

            raise TypeError(f"{self.classname()}: adjust_column should be a string")

        if len(mappings) > 1:

            if not isinstance(mappings, OrderedDict):

                raise TypeError(
                    f"{self.classname()}: mappings should be an ordered dict for 'replace' mappings using multiple columns"
                )

        self.adjust_column = adjust_column

    def transform(self, X):
        """Transforms values in given column using the values provided in the adjustments dictionary.

        Parameters
        ----------
        X : pd.DataFrame
            Data to apply adjustments to.

        Returns
        -------
        X : pd.DataFrame
            Transformed data X with adjustments applied to specified columns.

        """

        self.check_is_fitted(["adjust_column"])

        X = super().transform(X)

        if self.adjust_column not in X.columns.values:

            raise ValueError(
                f"{self.classname()}: variable {self.adjust_column} is not in X"
            )

        for i in self.columns:

            for j in self.mappings[i].keys():

                X[self.adjust_column] = np.where(
                    (X[i] == j), self.mappings[i][j], X[self.adjust_column]
                )

        return X


class CrossColumnMultiplyTransformer(BaseMappingTransformer):
    """Transformer to apply a multiplicative adjustment to values in one column based on the values of another column.

    Parameters
    ----------
    adjust_column : str
        The column to be adjusted.  The data type of this column must be int or float.

    mappings : dict
        Dictionary containing adjustments. Each value in adjustments should be a dictionary
        of key (column to apply adjustment based on) value (adjustment dict for given columns) pairs. For
        example the following dict {'a': {1: 2, 3: 5}, 'b': {'a': 0.5, 'b': 1.1}}
        would multiply the values in the adjustment column based off the values in column a using the mapping
        1->2*value, 3->5*value and also multiply based off the values in column b using a mapping
        'a'->0.5*value, 'b'->1.1*value.
        The values within the dicts defining the multipliers must have type int or float.

    **kwargs
        Arbitrary keyword arguments passed onto BaseTransformer.init method.

    Attributes
    ----------

    adjust_column : str
        Column containing the values to be adjusted.

    mappings : dict
        Dictionary of multiplicative adjustments for each column individually to be applied to the adjust_column.
        The dict passed to mappings in init is set to the mappings attribute.


    """

    def __init__(self, adjust_column, mappings, **kwargs):

        super().__init__(mappings=mappings, **kwargs)

        if not isinstance(adjust_column, str):

            raise TypeError(f"{self.classname()}: adjust_column should be a string")

        for j in mappings.values():

            for k in j.values():

                if type(k) not in [int, float]:

                    raise TypeError(
                        f"{self.classname()}: mapping values must be numeric"
                    )

        self.adjust_column = adjust_column

    def transform(self, X):
        """Transforms values in given column using the values provided in the adjustments dictionary.

        Parameters
        ----------
        X : pd.DataFrame
            Data to apply adjustments to.

        Returns
        -------
        X : pd.DataFrame
            Transformed data X with adjustments applied to specified columns.

        """

        self.check_is_fitted(["adjust_column"])

        X = super().transform(X)

        if self.adjust_column not in X.columns.values:

            raise ValueError(
                f"{self.classname()}: variable {self.adjust_column} is not in X"
            )

        if not pd.api.types.is_numeric_dtype(X[self.adjust_column]):

            raise TypeError(
                f"{self.classname()}: variable {self.adjust_column} must have numeric dtype."
            )

        for i in self.columns:

            for j in self.mappings[i].keys():

                X[self.adjust_column] = np.where(
                    (X[i] == j),
                    X[self.adjust_column] * self.mappings[i][j],
                    X[self.adjust_column],
                )

        return X


class CrossColumnAddTransformer(BaseMappingTransformer):
    """Transformer to apply an additive adjustment to values in one column based on the values of another column.

    Parameters
    ----------
    adjust_column : str
        The column to be adjusted.  The data type of this column must be int or float.

    mappings : dict
        Dictionary containing adjustments. Each value in adjustments should be a dictionary
        of key (column to apply adjustment based on) value (adjustment dict for given columns) pairs. For
        example the following dict {'a': {1: 2, 3: 5}, 'b': {'a': 1, 'b': -5}}
        would provide an additive adjustment to the values in the adjustment column based off the values
        in column a using the mapping 1->2+value, 3->5+value and also an additive adjustment based off the
        values in column b using a mapping 'a'->1+value, 'b'->(-5)+value.
        The values within the dicts defining the values to be added must have type int or float.

    **kwargs
        Arbitrary keyword arguments passed onto BaseTransformer.init method.

    Attributes
    ----------

    adjust_column : str
        Column containing the values to be adjusted.

    mappings : dict
        Dictionary of additive adjustments for each column individually to be applied to the adjust_column.
        The dict passed to mappings in init is set to the mappings attribute.


    """

    def __init__(self, adjust_column, mappings, **kwargs):

        super().__init__(mappings=mappings, **kwargs)

        if not isinstance(adjust_column, str):

            raise TypeError(f"{self.classname()}: adjust_column should be a string")

        for j in mappings.values():

            for k in j.values():

                if type(k) not in [int, float]:

                    raise TypeError(
                        f"{self.classname()}: mapping values must be numeric"
                    )

        self.adjust_column = adjust_column

    def transform(self, X):
        """Transforms values in given column using the values provided in the adjustments dictionary.

        Parameters
        ----------
        X : pd.DataFrame
            Data to apply adjustments to.

        Returns
        -------
        X : pd.DataFrame
            Transformed data X with adjustments applied to specified columns.

        """

        self.check_is_fitted(["adjust_column"])

        X = super().transform(X)

        if self.adjust_column not in X.columns.values:

            raise ValueError(
                f"{self.classname()}: variable " + self.adjust_column + " is not in X"
            )

        if not pd.api.types.is_numeric_dtype(X[self.adjust_column]):

            raise TypeError(
                f"{self.classname()}: variable "
                + self.adjust_column
                + " must have numeric dtype."
            )

        for i in self.columns:

            for j in self.mappings[i].keys():

                X[self.adjust_column] = np.where(
                    (X[i] == j),
                    X[self.adjust_column] + self.mappings[i][j],
                    X[self.adjust_column],
                )

        return X



"""
This module contains transformers that deal with imputation of missing values.
"""

import pandas as pd
import numpy as np
import warnings


from tubular.base import BaseTransformer


class BaseImputer(BaseTransformer):
    """Base imputer class containing standard transform method that will use pd.Series.fillna with the
    values in the impute_values_ attribute.

    Other imputers in this module should inherit from this class.
    """

    def transform(self, X):
        """Impute missing values with median values calculated from fit method.

        Parameters
        ----------
        X : pd.DataFrame
            Data to impute.

        Returns
        -------
        X : pd.DataFrame
            Transformed input X with nulls imputed with the median value for the specified columns.

        """

        self.check_is_fitted(["impute_values_"])

        X = super().transform(X)

        for c in self.columns:

            X[c] = X[c].fillna(self.impute_values_[c])

        return X


class ArbitraryImputer(BaseImputer):
    """Transformer to impute null values with an arbitrary pre-defined value.
    Parameters
    ----------
    impute_value : int or float or str
        Value to impute nulls with.
    columns : None or str or list, default = None
        Columns to impute, if the default of None is supplied all columns in X are used
        when the transform method is called.
    **kwargs
        Arbitrary keyword arguments passed onto BaseTransformer.init method.
    Attributes
    ----------
    impute_value : int or float or str
        Value to impute nulls with.
    """

    def __init__(self, impute_value, columns, **kwargs):

        if columns is None:

            raise ValueError(
                f"{self.classname()}: columns must be specified in init for ArbitraryImputer"
            )

        super().__init__(columns=columns, **kwargs)

        if (
            not isinstance(impute_value, int)
            and not isinstance(impute_value, float)
            and not isinstance(impute_value, str)
        ):

            raise ValueError(
                f"{self.classname()}: impute_value should be a single value (int, float or str)"
            )

        self.impute_values_ = {}
        self.impute_value = impute_value

    def transform(self, X):
        """Impute missing values with the supplied impute_value.
        If columns is None all columns in X will be imputed.
        Parameters
        ----------
        X : pd.DataFrame
            Data containing columns to impute.
        Returns
        -------
        X : pd.DataFrame
            Transformed input X with nulls imputed with the specified impute_value, for the specified columns.

        Additions
        ---------
        * Preserving the datatypes of columns
        * Finding the target column dtype and cast imputer values as same dtype
        """

        self.check_is_fitted(["impute_value"])
        self.columns_check(X)

        for c in self.columns:

            if "category" in X[c].dtype.name:

                if self.impute_value not in X[c].cat.categories:

                    X[c] = X[c].cat.add_categories(
                        self.impute_value
                    )  # add new category

            dtype = X[c].dtype  # get the dtype of column

            X[c] = (
                X[c].fillna(self.impute_values_).astype(dtype)
            )  # casting imputer value as same dtype

            self.impute_values_[
                c
            ] = self.impute_value  # updating impute_values_ attribute

        X = super().transform(X)  # impute the values

        return X


class MedianImputer(BaseImputer):
    """Transformer to impute missing values with the median of the supplied columns.

    Parameters
    ----------
    columns : None or str or list, default = None
        Columns to impute, if the default of None is supplied all columns in X are used
        when the transform method is called.

    weight: None or str, default=None
        Column containing weights

    **kwargs
        Arbitrary keyword arguments passed onto BaseTransformer.init method.

    Attributes
    ----------
    impute_values_ : dict
        Created during fit method. Dictionary of float / int (median) values of columns
        in the columns attribute. Keys of impute_values_ give the column names.

    """

    def __init__(self, columns=None, weight=None, **kwargs):

        super().__init__(columns=columns, **kwargs)

        if not isinstance(weight, str):
            if weight is not None:
                raise TypeError("weight should be str or None")

        self.weight = weight

    def fit(self, X, y=None):
        """Calculate median values to impute with from X.

        Parameters
        ----------
        X : pd.DataFrame
            Data to "learn" the median values from.

        y : None or pd.DataFrame or pd.Series, default = None
            Not required.

        """

        super().fit(X, y)

        self.impute_values_ = {}

        if self.weight is not None:

            super().check_weights_column(X, self.weight)

            temp = X.copy()

            for c in self.columns:

                # filter out null rows so their weight doesn't influence calc
                filtered = temp[temp[c].notnull()]

                # first sort df by column to be imputed (order of weight column shouldn't matter for median)
                filtered.sort_values(c, inplace=True)

                # next calculate cumulative weight sums
                cumsum = filtered[self.weight].cumsum()

                # find midpoint
                cutoff = filtered[self.weight].sum() / 2.0

                # find first value >= this point
                median = filtered[c][cumsum >= cutoff].iloc[0]

                self.impute_values_[c] = median

        else:

            for c in self.columns:

                self.impute_values_[c] = X[c].median()

        return self


class MeanImputer(BaseImputer):
    """Transformer to impute missing values with the mean of the supplied columns.

    Parameters
    ----------
    columns : None or str or list, default = None
        Columns to impute, if the default of None is supplied all columns in X are used
        when the transform method is called.

    weights : None or str, default = None
        Column containing weights.

    **kwargs
        Arbitrary keyword arguments passed onto BaseTransformer.init method.

    Attributes
    ----------
    impute_values_ : dict
        Created during fit method. Dictionary of float / int (mean) values of columns
        in the columns attribute. Keys of impute_values_ give the column names.

    """

    def __init__(self, columns=None, weight=None, **kwargs):

        super().__init__(columns=columns, **kwargs)

        if not isinstance(weight, str):
            if weight is not None:
                raise TypeError("weight should be str or None")

        self.weight = weight

    def fit(self, X, y=None):
        """Calculate mean values to impute with from X.

        Parameters
        ----------
        X : pd.DataFrame
            Data to "learn" the mean values from.

        y : None or pd.DataFrame or pd.Series, default = None
            Not required.

        """

        super().fit(X, y)

        self.impute_values_ = {}

        if self.weight is not None:

            super().check_weights_column(X, self.weight)

            for c in self.columns:

                # filter out null rows so they don't count towards total weight
                filtered = X[X[c].notnull()]

                # calculate total weight and total of weighted col
                total_weight = filtered[self.weight].sum()
                total_weighted_col = filtered[c].mul(filtered[self.weight]).sum()

                # find weighted mean and add to dict
                weighted_mean = total_weighted_col / total_weight

                self.impute_values_[c] = weighted_mean

        else:

            for c in self.columns:

                self.impute_values_[c] = X[c].mean()

        return self


class ModeImputer(BaseImputer):
    """Transformer to impute missing values with the mode of the supplied columns.

    If mode is NaN, a warning will be raised.

    Parameters
    ----------
    columns : None or str or list, default = None
        Columns to impute, if the default of None is supplied all columns in X are used
        when the transform method is called.

    weight : str
        Name of weights columns to use if mode should be in terms of sum of weights
        not count of rows.

    **kwargs
        Arbitrary keyword arguments passed onto BaseTransformer.init method.

    Attributes
    ----------
    impute_values_ : dict
        Created during fit method. Dictionary of float / int (mode) values of columns
        in the columns attribute. Keys of impute_values_ give the column names.

    """

    def __init__(self, columns=None, weight=None, **kwargs):

        super().__init__(columns=columns, **kwargs)

        if weight is not None:

            if not isinstance(weight, str):

                raise ValueError("ModeImputer: weight should be a string or None")

        self.weight = weight

    def fit(self, X, y=None):
        """Calculate mode values to impute with from X.

        Parameters
        ----------
        X : pd.DataFrame
            Data to "learn" the mode values from.

        y : None or pd.DataFrame or pd.Series, default = None
            Not required.

        """

        super().fit(X, y)

        self.impute_values_ = {}

        if self.weight is None:

            for c in self.columns:

                mode_value = X[c].mode(dropna=True)

                if len(mode_value) == 0:

                    self.impute_values_[c] = np.nan

                    warnings.warn(f"ModeImputer: The Mode of column {c} is NaN.")

                else:

                    self.impute_values_[c] = mode_value[0]

        else:

            super().check_weights_column(X, self.weight)

            for c in self.columns:

                self.impute_values_[c] = X.groupby(c)[self.weight].sum().idxmax()

        return self


class NearestMeanResponseImputer(BaseImputer):
    """Class to impute missing values with; the value for which the average response is closest
    to the average response for the unknown levels.

    Parameters
    ----------
    columns : None or str or list, default = None
        Columns to impute, if the default of None is supplied all columns in X are used
        when the transform method is called.

    """

    def __init__(self, columns=None, **kwds):

        super().__init__(columns=columns, **kwds)

    def fit(self, X, y):
        """Calculate mean values to impute with.

        Parameters
        ----------
        X : pd.DataFrame
            Data to fit the transformer on.

        y : pd.Series
            Response column used to determine the value to impute with. The average response for
            each level of every column is calculated. The level which has the closest average response
            to the average response of the unknown levels is selected as the imputation value.

        """

        super().fit(X, y)

        n_nulls = y.isnull().sum()

        if n_nulls > 0:

            raise ValueError(f"{self.classname()}: y has {n_nulls} null values")

        self.impute_values_ = {}

        X_y = self._combine_X_y(X, y)
        response_column = "_temporary_response"

        for c in self.columns:

            c_nulls = X[c].isnull()

            if c_nulls.sum() == 0:

                raise ValueError(
                    f"{self.classname()}: Column {c} has no missing values, cannot use this transformer."
                )

            else:

                mean_response_by_levels = pd.DataFrame(
                    X_y.loc[~c_nulls].groupby(c)[response_column].mean()
                ).reset_index()

                mean_response_nulls = X_y.loc[c_nulls, response_column].mean()

                mean_response_by_levels["abs_diff_response"] = np.abs(
                    mean_response_by_levels[response_column] - mean_response_nulls
                )

                # take first value having the minimum difference in terms of average response
                self.impute_values_[c] = mean_response_by_levels.loc[
                    mean_response_by_levels["abs_diff_response"]
                    == mean_response_by_levels["abs_diff_response"].min(),
                    c,
                ].values[0]

        return self


class NullIndicator(BaseTransformer):
    """Class to create a binary indicator column for null values.

    Parameters
    ----------
    columns : None or str or list, default = None
        Columns to produce indicator columns for, if the default of None is supplied all columns in X are used
        when the transform method is called.

    """

    def __init__(self, columns=None, **kwds):

        super().__init__(columns=columns, **kwds)

    def transform(self, X):
        """Create new columns indicating the position of null values for each variable in self.columns.

        Parameters
        ----------
        X : pd.DataFrame
            Data to add indicators to.

        """

        X = super().transform(X)

        for c in self.columns:

            X[f"{c}_nulls"] = X[c].isnull().astype(int)

        return X



import setuptools
import re

with open("README.md", "r") as fh:
    long_description = fh.read()

# get version from _version.py file, from below
# https://stackoverflow.com/questions/458550/standard-way-to-embed-version-into-python-package
VERSION_FILE = "tubular/_version.py"
version_file_str = open(VERSION_FILE, "rt").read()
VERSION_STR_RE = r"^__version__ = ['\"]([^'\"]*)['\"]"
mo = re.search(VERSION_STR_RE, version_file_str, re.M)
if mo:
    version = mo.group(1)
else:
    raise RuntimeError("Unable to find version string in %s." % (VERSION_FILE,))


def list_reqs(fname="requirements.txt"):
    with open(fname) as fd:
        return fd.read().splitlines()


setuptools.setup(
    name="tubular",
    version=version,
    author="LV GI Data Science Team",
    author_email="#DataSciencePackages@lv.co.uk",
    description="Package to perform pre processing steps for machine learning models",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),
    install_requires=list_reqs(),
    python_requires=">=3.6",
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Operating System :: OS Independent",
        "License :: OSI Approved :: BSD License",
    ],
)



from tubular.base import BaseTransformer
import pandas as pd


class SetValueTransformer(BaseTransformer):
    """Transformer to set value of column(s) to a given value.

    This should be used if columns need to be set to a constant value.

    Parameters
    ----------
    columns: list or str
        Columns to set values.

    value : various
        Value to set.

    **kwargs
        Arbitrary keyword arguments passed onto BaseTransformer.init method.

    """

    def __init__(self, columns, value, **kwargs):

        self.value = value

        super().__init__(columns=columns, **kwargs)

    def transform(self, X):
        """Set columns to value.

        Parameters
        ----------
        X : pd.DataFrame
            Data to apply mappings to.

        Returns
        -------
        X : pd.DataFrame
            Transformed input X with columns set to value.

        """

        X = super().transform(X)

        X[self.columns] = self.value

        return X


class SetColumnDtype(BaseTransformer):
    """
    Transformer to set transform columns in a dataframe to a dtype

    Parameters
    ----------
    columns : str or list
        Columns to set dtype. Must be set or transform will not run.

    dtype : type or string
        dtype object to set columns to or a string interpretable as one by pd.api.types.pandas_dtype
        e.g. float or 'float'
    """

    def __init__(self, columns, dtype):

        super().__init__(columns, copy=True)

        self.__validate_dtype(dtype)

        self.dtype = dtype

    def transform(self, X):

        X = super().transform(X)

        X[self.columns] = X[self.columns].astype(self.dtype)

        return X

    def __validate_dtype(self, dtype: str):
        """Check string is a valid dtype"""

        try:
            pd.api.types.pandas_dtype(dtype)
        except TypeError:
            raise TypeError(
                f"{self.classname()}: data type '{dtype}' not understood as a valid dtype"
            )
